local Sackville = LibStub("AceAddon-3.0"):GetAddon("Sackville", "AceConsole-3.0")
local B = LibStub("LibBabble-Inventory-3.0")
local BL = B:GetLookupTable()

-- These are "standard" because they match up with the Type and SubType
-- values returned by GetItemInfo(itemid)
local standardCategories = {
    [BL["Armor"]] = {
        BL["Miscellaneous"],
        BL["Cloth"],
        BL["Leather"],
        BL["Mail"],
        BL["Plate"],
        BL["Shields"],
        BL["Librams"],
        BL["Idols"],
        BL["Totems"],
        BL["Sigils"]
    },
    [BL["Consumable"]] = {
        BL["Food & Drink"],
        BL["Potion"],
        BL["Elixir"],
        BL["Flask"],
        BL["Bandage"],
        BL["Item Enhancement"],
        BL["Scroll"],
        BL["Other"]
    },
    [BL["Container"]] = {
        BL["Bag"],
        BL["Enchanting Bag"],
        BL["Engineering Bag"],
        BL["Gem Bag"],
        BL["Herb Bag"],
        BL["Mining Bag"],
        BL["Soul Bag"],
        BL["Leatherworking Bag"]
    },
    [BL["Gem"]] = {
        BL["Blue"],
        BL["Green"],
        BL["Orange"],
        BL["Meta"],
        BL["Prismatic"],
        BL["Purple"],
        BL["Red"],
        BL["Simple"],
        BL["Yellow"]
    },
    [BL["Miscellaneous"]] = {
        BL["Junk"],
        BL["Reagent"],
        BL["Pet"],
        BL["Holiday"],
        BL["Mount"],
        BL["Other"]
    },
    [BL["Recipe"]] = {
        BL["Alchemy"],
        BL["Blacksmithing"],
        BL["Book"],
        BL["Cooking"],
        BL["Enchanting"],
        BL["Engineering"],
        BL["First Aid"],
        BL["Leatherworking"],
        BL["Tailoring"]
    },
    [BL["Quest"]] = {
        BL["Quest"]
    },
    [BL["Trade Goods"]] = {
        BL["Armor Enchantment"],
        BL["Cloth"],
        BL["Cooking"],
        BL["Devices"],
        BL["Elemental"],
        BL["Enchanting"],
        BL["Explosives"],
        BL["Herb"],
        BL["Jewelcrafting"],
        BL["Leather"],
        BL["Materials"],
        BL["Meat"],
        BL["Metal & Stone"],
        BL["Other"],
        BL["Parts"],
        BL["Trade Goods"],
        BL["Weapon Enchantment"]
    },
    [BL["Weapon"]] = {
        BL["Bows"],
        BL["Crossbows"],
        BL["Daggers"],
        BL["Guns"],
        BL["Fishing Poles"],
        BL["Fist Weapons"],
        BL["Miscellaneous"],
        BL["One-Handed Axes"],
        BL["One-Handed Maces"],
        BL["One-Handed Swords"],
        BL["Polearms"],
        BL["Staves"],
        BL["Thrown"],
        BL["Two-Handed Axes"],
        BL["Two-Handed Maces"],
        BL["Two-Handed Swords"],
        BL["Wands"]
    }
}

-- Create and return a function that compares an item based on its
-- itemType and itemSubType as returned by GetItemInfo(itemid)
local function MakeStandardFilter(itemType, itemSubType)
    assert(itemType ~= nil)

    local code = "return function (item) "    
    if itemSubType then
        code = code .. string.format([[
            return (item.type == "%s" and item.subType == "%s")]],
            itemType, itemSubType)
    else
        code = code .. string.format([[
            return (item.type == "%s")]], itemType)
    end
    
    code = code .. " end"
    return assert(loadstring(code)())
end

for itemType, itemSubTypes in pairs(standardCategories) do
    local typeFilter = MakeStandardFilter(itemType)
    Sackville.Filter:AddFilter(itemType, typeFilter)
    
    for _, itemSubType in ipairs(itemSubTypes) do
        local subTypeFilter = MakeStandardFilter(itemType, itemSubType)
        Sackville.Filter:AddFilter(itemType .. "\001" .. itemSubType, subTypeFilter)
    end
end