Sackville = LibStub("AceAddon-3.0"):NewAddon("Sackville", "AceConsole-3.0", "AceHook-3.0", 'AceEvent-3.0')
local L = LibStub("AceLocale-3.0"):GetLocale("Sackville")
Sackville.version = "Sackville v0.5"

local sprintf = string.format
local _
local debugOpts = {
    main        = 0,
    event       = 0,
    bag         = 0,
    section     = 0,
    bagLayout   = 0,
    filter      = 0,
    item        = 0,
    empty       = 0
}

local bags = {}
local quickPanel = nil

local bagsAreOpen = false
local bankBagsInitialized = false
local playerBagsInitialized = false
local isBankOpen = false
local bagIndexes = {}

local function BuildBagIndexes()
   bagIndexes = {
      player = {},
      bank = {}
   }

   for i = 0, NUM_BAG_SLOTS do
      table.insert(bagIndexes.player, i)
   end
   
   table.insert(bagIndexes.bank, BANK_CONTAINER)
   for i = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
      table.insert(bagIndexes.bank, i)
   end
end

local function PositionFrame(bagFrame, state)
    -- local bagFrame = bag:GetFrame()
    local h = bagFrame:GetHeight()
    local w = bagFrame:GetWidth()
    Sackville:Debug(1, "bagLayout",
        " PositionFrame, w: %d, h: %d, nextPos:(%d,%d), widest: %d, SH: %d",
        w, h, state.x_offset, state.y_offset, state.widest_bag, GetScreenHeight())
    state.widest_bag = (w > state.widest_bag) and w or state.widest_bag
    
    if (state.y_offset + h) < GetScreenHeight() then
        -- There is enough room to stack this frame vertically
        Sackville:Debug(1, "bagLayout", "  Setting pos to (%d, %d)",
            state.x_offset, state.y_offset)
        bagFrame:SetPoint("BOTTOMRIGHT", state.x_offset, state.y_offset)
        state.y_offset = state.y_offset + h + Sackville.Options.FRAME_PADDING
    else
        -- Start a new column
        state.y_offset = Sackville.Options.layout.origin.y_offset
        state.x_offset = state.x_offset - state.widest_bag - Sackville.Options.FRAME_PADDING
        state.widest_bag = 0
        Sackville:Debug(1, "bagLayout", "  Setting pos to (%d, %d)", 
            state.x_offset, state.y_offset)
        bagFrame:SetPoint("BOTTOMRIGHT", state.x_offset, state.y_offset)
        state.y_offset = state.y_offset + h + Sackville.Options.FRAME_PADDING
    end
    
    return state
end

-- Event handlers --

function Sackville:OnInitialize()
-- Code that you want to run when the addon is first loaded goes here.
-- do init tasks here, like restoring saved variables, 
-- or setting up slash commands.

    self.db = LibStub("AceDB-3.0"):New("SackvilleDB", Sackville.Options.defaults)
    self.db:ResetDB() -- TODO: Temporary until the config GUI is working
    
    -- db = self.db.profile
    Sackville:RegisterChatCommand("sack", "ChatCommand")
    self.dummyFrame = CreateFrame("Frame", "SackvilleDummyFrame", UIParent)
    self.dummyFrame:Hide()
    quickPanel = Sackville.Bag:CreateQuickPanel()
    BuildBagIndexes()
    self:BuildVirtualBags()
    
    for bagslot, item in pairs(self.db.char.items) do
        setmetatable(item, Sackville.Item)
        item:RecreateButton()
        item:ReAddToSection()
        -- item:SetEmpty(true) -- Forces items to be reprocessed
    end
end

function Sackville:OnEnable()
-- Do more initialization here, that really enables the use of your addon.
-- Register Events, Hook functions, Create Frames, Get information from 
-- the game that wasn't available in OnInitialize
    -- self:RegisterEvent('EQUIPMENT_SETS_CHANGED')
    if self.enabled then return; end
    self:RegisterEvent("BANKFRAME_OPENED")
    self:RegisterEvent("BANKFRAME_CLOSED")
    self:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
    self:RegisterEvent('AUCTION_HOUSE_SHOW', "ShowPlayerBags")
    self:RegisterEvent('AUCTION_HOUSE_CLOSED', "HidePlayerBags")
    self:RegisterEvent("PLAYER_MONEY", "UpdateMoney")
    self:RegisterEvent("CURRENCY_DISPLAY_UPDATE", "UpdateMoney")

    self:RawHook("ToggleAllBags", "TogglePlayerBags", true)
    self:RawHook("ToggleBackpack", "TogglePlayerBags", true)
    self:RawHook("ToggleBag", "TogglePlayerBags", true)
    self:RawHook("OpenAllBags", "ShowPlayerBags", true)
    self:RawHook("OpenBackpack", "ShowPlayerBags", true)
    self:RawHook("CloseAllWindows", "CloseAllWindows", true)
    self:RawHookScript(BankFrame, "OnEvent", "DoNothing")
    -- self:RawHook("CloseBackpack", "HidePlayerBags", true)
    -- self:RawHook("SetCurrencyBackpack", "SetCurrencyBackpack", true)
    
    self.enabled = true
end

function Sackville:OnDisable()
    if not self.enabled then return; end
    self:UnregisterEvent("BANKFRAME_OPENED")
    self:UnregisterEvent("BANKFRAME_CLOSED")
    self:UnregisterEvent("PLAYERBANKSLOTS_CHANGED")
    self:UnregisterEvent('AUCTION_HOUSE_SHOW', "ShowPlayerBags")
    self:UnregisterEvent('AUCTION_HOUSE_CLOSED', "HidePlayerBags")
    self:UnregisterEvent("PLAYER_MONEY", "UpdateMoney")
    self:UnregisterEvent("CURRENCY_DISPLAY_UPDATE", "UpdateMoney")
    
    self:UnhookAll()

    self:HidePlayerBags()
    self:HideBankBags()
    if isBankOpen then
        CloseBankFrame();
    end
    self.enabled = false
end

function Sackville:BAG_UPDATE(event, bagid)
    self:Debug(1, "event", "BAG_UPDATE: %s", bagid)

    local bagType

    if self:IsPlayerBagId(bagid) then
        bagType = PLAYER
        if (not playerBagsInitialized) then self:InitializePlayerBags(); return; end
    elseif (isBankOpen and self:IsBankBagId(bagid)) then
        bagType = L["Bank"]
        if (not bankBagsInitialized) then self:InitializeBankBags(); return; end
    end

    local result = self:RebuildInventory(bagid)
    self:ProcessScanResult(result, bagType)

    -- Rebuild any bags that potentially had items added/changed/removed, and
    -- then layout the bags if it is deemed necessary.
    for index, bag in self:GetOrderedBags(bagType) do
        if bag:LayoutItems() then self:LayoutBags(bag) end
    end
    
    -- TODO: Don't reposition bag frames unless we absolutely need to.
    -- if result then 
        -- self:LayoutBags()
    -- end
end

function Sackville:BANKFRAME_OPENED()
    self:Debug(1, "event", "BANKFRAME_OPENED")
    isBankOpen = true
    if not bankBagsInitialized then self:InitializeBankBags() end
    self:ShowPlayerBags()
    self:ShowBankBags()
    self:LayoutBags()
end

function Sackville:BANKFRAME_CLOSED()
    self:Debug(1, "event", "BANKFRAME_CLOSED")
    self:HideBankBags()
    isBankOpen = false       
end

function Sackville:PLAYERBANKSLOTS_CHANGED(event, arg1)
    self:Debug(1, "event", "PLAYERBANKSLOTS_CHANGED: %s", arg1)
    -- When (arg1 <= NUM_BANKGENERIC_SLOTS), arg1 is the index of the generic
    -- bank slot that changed. When (arg1 > NUM_BANKGENERIC_SLOTS), 
    -- (arg1 - NUM_BANKGENERIC_SLOTS) is the index of the equipped bank bag that changed. 
    local bagid
    if (arg1 <= NUM_BANKGENERIC_SLOTS) then -- generic bank slot changed
        self:BAG_UPDATE(event, -1)
    else
        -- bagid = arg1 - NUM_BANKGENERIC_SLOTS    
        -- self:BAG_UPDATE(event, bagid)
    end
end

function Sackville:BAG_UPDATE_COOLDOWN(event, cid)
    cid = cid or "?"
    self:Debug(1, "event", "BAG_UPDATE_COOLDOWN:", (cid or "?"))
    
    for _, item in pairs(self.db.char.items) do
        if item and not item.empty then item:UpdateButtonCooldown() end
    end

end

function Sackville:ITEM_LOCK_CHANGED(event, bagid, slot)
-- Desaturate item button when it is locked (probably because you're moving it)
    self:Debug(1, "event", "ITEM_LOCK_CHANGED")
    if not (bagid and slot) then return end
    local item = self.db.char.items[bagid .. '-' .. slot]
    if (item and not item:IsEmpty()) then
        local _, _, locked = GetContainerItemInfo(bagid, slot);
        item:SetButtonDesaturated(locked)
    end

    -- FIXME: If the bank bag is closed while the item is being dragged,
    -- the button will stay desaturated (but not locked)
end

function Sackville:SearchBoxUpdate(searchBox, userChanged)
    local text = searchBox:GetText();
    if ( text == SEARCH ) then
        text = "";
    end
    HighlightSearchItems(text)
    if (text ~= "") then
        searchBox.clearButton:Show();
    else
        searchBox.clearButton:Hide();
    end
end
-- Inventory processing --

function Sackville:BuildVirtualBags()
    for bagName, bagOptions in pairs(self.db.profile.bagConfig) do
        if not bags[bagName] then
            local bag = Sackville.Bag:New(bagName, bagOptions)
            bags[bagName] = bag

            for sectionName, sectionOptions in pairs (bagOptions.sections) do        
                bag:AddSection(sectionName, sectionOptions)
            end

            self:Debug(1, "main", "Initialized virtual bag (%s)", bag.name)
        end
    end
end

function Sackville:InitializePlayerBags()
    self:Debug(1, "main", "InitializePlayerBags")
    -- Scan the player bags and take appropriate action for any items that
    -- have been changed/added/removed from what we had saved (if anything)
    for _, bagid in ipairs(bagIndexes.player) do
        local result = self:RebuildInventory(bagid)           
        self:ProcessScanResult(result, PLAYER)
    end
    playerBagsInitialized = true
    
    -- Get the bag contents arranged and the bag properly sized
    for index, bag in self:GetOrderedBags("Player") do
        if bag:LayoutItems() then self:LayoutBags() end
    end
end

function Sackville:InitializeBankBags()
    self:Debug(1, "main", "InitializeBankBags")
    -- Scan the player bags and take appropriate action for any items that
    -- have been changed/added/removed from what we had saved (if anything)    
    for _, bagid in ipairs(bagIndexes.bank) do
        local result = self:RebuildInventory(bagid)           
        self:ProcessScanResult(result, L["Bank"])
    end
    bankBagsInitialized = true
    
    -- Get the bag contents arranged and the bag properly sized
    for index, bag in self:GetOrderedBags(L["Bank"]) do
        bag:LayoutItems()
    end
end

function Sackville:IsBagType(bagType, bagid)
    if bagType == PLAYER and self:IsPlayerBagId(bagid) then
        return true
    elseif bagType == L["Bank"] and self:IsBankBagId(bagid) then
        return true
    else
        return false
    end
end

function Sackville:IsBankBagId(bagId)
    for _, bankBagId in ipairs(bagIndexes.bank) do
        if bagId == bankBagId then
            return true
        end
    end
    
    return false
end

function Sackville:IsPlayerBagId(bagId)
    for _, playerBagId in ipairs(bagIndexes.player) do
        if bagId == playerBagId then
            return true
        end
    end
    
    return false
end

function Sackville:RebuildInventory(updatedBagId)
-- Scans a specific bag (or all bags if none specified) and 
-- populates/updates self.db.char.items, returning a table of all items that
-- have been added/changed/deleted.
    local bagid = updatedBagId or "nil"
    self:Debug(2, "main", "RebuildInventory: %s", bagid)
    local result = nil    
    local bagType = nil
    
    if updatedBagId ~= nil then
        if self:IsPlayerBagId(updatedBagId) then
            bagType = PLAYER
        elseif self:IsBankBagId(updatedBagId) then
            bagType = L["Bank"]
        else
            assert(bagType, "Unknown bag type for bagid: %s ", bagid)
            -- Support future bag types (guild vault, void storage)
        end

        if bagType ~= nil then -- Scan a specific bag
            local slots = GetContainerNumSlots(updatedBagId)
            if slots then
                result = Sackville:ScanBagSlots(updatedBagId, slots, result)
            else
                self:Debug(2, "main", "Bag %s has no slots", bagid)
            end
        end
    else -- Scan everything
        for _, bagid in ipairs(bagIndexes.player) do
            local slots = GetContainerNumSlots(bagid)
            if slots then
                result = Sackville:ScanBagSlots(bagid, slots, result)
            else
                self:Debug(2, "main", "Player bag %s has no slots", bagid)
            end
        end
        
        if isBankOpen then
            for _, bagid in ipairs(bagIndexes.bank) do
                local slots = GetContainerNumSlots(bagid)
                if slots then
                    result = Sackville:ScanBagSlots(bagid, slots, result)
                else
                    self:Debug(2, "main", "Bank bag %s has no slots", bagid)
                end
            end
        end
    end
    
    return result
end

function Sackville:ScanBagSlots(bagid, slots, priorResult)
    self:Debug(2, "main", "  ScanBagSlots (" .. bagid .. ", " .. slots .. ")")

    local result = {}
    
    if priorResult then result = priorResult
    else
        result.itemsAdded = {}
        result.itemsChanged = {}
        result.itemsRemoved = {}    
    end
    
    for slot = 1, slots do
        local bagslot = sprintf("%d-%d", bagid, slot)
        local item = self.db.char.items[bagslot]
        if item then
            local itemStatus = item:Reload()
            if itemStatus == "removed" then
                table.insert(result.itemsRemoved, self.db.char.items[bagslot])
            elseif itemStatus == "added" then 
                table.insert(result.itemsAdded, self.db.char.items[bagslot])           
            elseif itemStatus == "changed" then
                table.insert(result.itemsChanged, self.db.char.items[bagslot])        
            end
        else
            item = Sackville.Item:New(bagid, slot)
            self.db.char.items[bagslot] = item
            if item:IsEmpty() then
                self:Debug(4, "main", "Skipping empty slot (%d,%d)", bagid, slot)
            else
                table.insert(result.itemsAdded, item)
            end
        end
    end
    
    self:Debug(2, "main", "  # of items added: %d, changed: %d, removed: %d",
        #result.itemsAdded, #result.itemsChanged, #result.itemsRemoved)
    return result
end

function Sackville:ProcessScanResult(result, bagType)
    self:Debug(2, "main", "ProcessScanResult, %s", bagType)
    for _, item in ipairs(result.itemsAdded) do
        local section  = self:GetBagSectionForItem(item, bagType)
        if section then
            section:AddItem(item)
        else
            Sackville:Print("Couldn't find a bag-section for (%s)", item:Dump())
        end            
    end
    
    for _, item in ipairs(result.itemsRemoved) do 
        item:Release()
    end
    
    for _, item in ipairs(result.itemsChanged) do
        item:Show()
        local section = self:GetBagSectionForItem(item, bagType)
        if section then            
            -- Should this item be reassigned to a different section?
            if section.bagName ~= item.bagName
            or (section.name ~= item.sectionName) then
                item:Release()
                -- bags[item.bagName].sections[item.sectionName]:RemoveItem(item)
                section:AddItem(item)
            end
        else
            self:Debug(1, "item", "Couldn't find a bag for %s", item.name)
        end            

    end
end

function Sackville:GetBagSectionForItem(item, bagType)
-- item: The item we're trying to find a matching section for
-- bagType: which bag type we're allowed to match (player, bank, ...)
    self:Debug(1, "item", "GetBagSectionForItem: %s", item.name)
    bagType = bagType or item:GetBagType()
    
    local matches = {}
    bagSet = self:GetBagSet(bagType)
        
    for bagName, bag in pairs(bagSet) do
        for sectionName, section in pairs(bag.sections) do
            
            if section:MatchItem(item) then
                self:Debug(3, "item", "> Item: %s, found matching section: %s-%s with prio: %d",
                item.name, bag.name, section.name, section.priority)
                table.insert(matches, section)
            end
        end

    end
    
    -- Sort by descending priority. Highest priority wins.
    for i, match in pairs(matches) do
        self:Debug(3, "item", "- Matching section: %s, prio: %d, index: %d",
            match.name, match.priority, i)
    end
    local matchComparer = function(a, b) return a.priority > b.priority end
    table.sort(matches, matchComparer)
    
    return matches[1]
end

-- Bag utility --

function Sackville:GetBagSet(bagType)
    local bagSet = {}
    for bagName, bag in pairs(bags) do
        if bag.bagType == bagType then bagSet[bagName] = bag end
    end
    return bagSet
end

function Sackville:GetOrderedBags(bagType)   
    local bagSortOptions = {
        {
            key = "displayOrder",
            comparison = "<"
        }
    }
    
    local bagSet
    if bagType then
        bagSet = self:GetBagSet(bagType)
    else
        bagSet = bags
    end
    
    return self.Util.kpairs(bagSet, bagSortOptions)
end

function Sackville:UpdateMoney()
    quickPanel.money:SetText(Sackville.GetMoneyString())
    w = quickPanel.money:GetStringWidth() + Sackville.Options.TITLE_EXTRA_WIDTH
    quickPanel:SetWidth(w)
    quickPanel.titleFrame:SetWidth(w)
end

function Sackville:GetMoneyString()
    local copper = GetMoney()
    local money

    local size = 12
    local text = sprintf('%d|TInterface\\MoneyFrame\\UI-GoldIcon:%d:%d:3:0:64:64:4:60:4:60|t ', 
        copper / 100 / 100, size, size)
    text = text .. sprintf(' %d|TInterface\\MoneyFrame\\UI-SilverIcon:%d:%d:3:0:64:64:4:60:4:60|t ', 
        (copper / 100) % 100, size, size)
    text = text .. sprintf(' %d|TInterface\\MoneyFrame\\UI-CopperIcon:%d:%d:3:0:64:64:4:60:4:60|t ', 
        copper % 100, size, size)

    for i=1, GetCurrencyListSize() do
        local name, isHeader, _, _, isWatched, count, icon = GetCurrencyListInfo(i)
        if (not isHeader) then
            if isWatched then
                text = (count or 0) .. '|T' .. icon .. ':' .. size .. ':' .. size
                    .. ':3:0:64:64:4:60:4:60|t   ' .. text
            end
            if money then
                money[i] = count
            end
        end
    end

   return text
end

-- Functions directly related to bag display/layout --

function Sackville:LayoutBags(startingBag)
    self:Debug(1, "main", "LayoutBags()")
    startingBagIndex = startingBag and startingBag.displayOrder or 1
    
    local positionState = {
        x_offset = Sackville.Options.layout.origin.x_offset,
        y_offset = Sackville.Options.layout.origin.y_offset,
        widest_bag = 0
    }
    
    local bagSortOptions = {
        {
            key = "displayOrder",
            comparison = "<"
        }
    }
    
    positionState = PositionFrame(quickPanel, positionState)
    
    for index, bag in self:GetOrderedBags(PLAYER) do
        -- if index >= startingBagIndex then
            positionState = PositionFrame(bag:GetFrame(), positionState)
        -- end
    end
    
    if isBankOpen then
        for index, bag in self:GetOrderedBags(L["Bank"]) do
            -- bag:LayoutItems()
            positionState = PositionFrame(bag:GetFrame(), positionState)
        end
    end
    quickPanel:Show()
end

function Sackville:HideBags()
    self:CloseAllWindows()
end

function Sackville:CloseAllWindows()
    self:Debug(1, "event", "CloseAllWindows")
    local bagsWereOpen = bagsAreOpen
    if bagsAreOpen then
        self:HidePlayerBags()
    end
    if isBankOpen then
        CloseBankFrame();
    end
    windowsVisible = CloseWindows(ignoreCenter);
    Sackville:ClearSearchBox()
    return bagsWereOpen or windowsVisible
end

function Sackville:TogglePlayerBags()
    if bagsAreOpen then
        self:HidePlayerBags()
    else
        self:ShowPlayerBags()
    end
end

function Sackville:ShowPlayerBags()
    self:Debug(1, "main", "ShowPlayerBags()")
    if ( UnitIsDead(PLAYER) ) then
        NotWhileDeadError();
        return
    end
    
    if not playerBagsInitialized then self:InitializePlayerBags() end   
    
    Sackville:RegisterEvent("BAG_UPDATE")
	Sackville:RegisterEvent("ITEM_LOCK_CHANGED")
	Sackville:RegisterEvent("BAG_UPDATE_COOLDOWN")

    self:LayoutBags()
    
    quickPanel:Show()
    self:UpdateMoney()
      
    for _, bag in pairs(self:GetBagSet(PLAYER)) do
        bag:Show()
    end
    
    bagsAreOpen = true
    PlaySound("igBackPackOpen")
end

function Sackville:HidePlayerBags()
    self:Debug(1, "main", "HidePlayerBags()")

	Sackville:UnregisterEvent("BAG_UPDATE")
	Sackville:UnregisterEvent("ITEM_LOCK_CHANGED")
	Sackville:UnregisterEvent("BAG_UPDATE_COOLDOWN")
    
    for _, bag in pairs(self:GetBagSet(PLAYER)) do
        bag:Hide()
    end

    quickPanel:Hide()
    PlaySound("igBackPackClose")
    
    if bagsAreOpen then
        bagsAreOpen = false
        return true
    else
        bagsAreOpen = false
        return false
    end
    
    Sackville:ClearSearchBox()
end

function Sackville:ShowBankBags()
    for _, bag in pairs(self:GetBagSet(L["Bank"])) do
        bag:Show()
    end
end

function Sackville:HideBankBags()
    for _, bag in pairs(self:GetBagSet(L["Bank"])) do
        bag:Hide()
    end
end

-- Miscellaneous --

function HighlightSearchItems(text)
    for bagName, bag in pairs(bags) do
        bag:HighlightSearchItems(text)
    end
end

function Sackville:ClearSearchBox()
    local editBox = quickPanel.searchBox
    if editBox.clearFunc then
        editBox.clearFunc(editBox);
    end
    
    editBox:SetText("");
    if not editBox:HasFocus() then
        editBox:GetScript("OnEditFocusLost")(editBox);
    end
    editBox:ClearFocus();
end

function Sackville:DoNothing()
end

function Sackville:GetEmptySlotCount(bagType)
    local emptyCount = 0
    local bagSet
    
    if bagType == PLAYER then
        bagSet = bagIndexes.player
    elseif bagType == L["Bank"] then
        bagSet = bagIndexes.bank
    end
    
    if bagSet then
        for _, bagId in ipairs(bagSet) do
            numberOfFreeSlots, unused = GetContainerNumFreeSlots(bagId);
            emptyCount = emptyCount + numberOfFreeSlots
        end
    end
    
    return emptyCount
end

function Sackville:ChatCommand(input)
    input = (input) and input or ""

    if not input or input:trim() == '' then
        self:TogglePlayerBags()
    elseif input == 'bi' then
        self:RebuildInventory()
    elseif input == 'lay' then
        self:LayoutBags()
    elseif input == "ct" then
        self.Config:CreateGUI()
    elseif input == "rdb" then
        self.db:ResetDB()
    elseif input == "enable" then
        self:OnEnable()
    elseif input == "disable" then
        self:OnDisable()
    end
end

function Sackville:Debug(verbosity, flag, formatString, ...)
    if debugOpts[flag] >= verbosity then
        self:Print(sprintf(formatString, ...))
    end
end

function Sackville:GetBag(bagName)
    return bags[bagName]
end

function Sackville:IsBankOpen()
    return isBankOpen
end